from __future__ import annotations
from dataclasses import dataclass
from typing import cast, Optional, TypeVar
from typing import Tuple

from relationalai.early_access.metamodel import builtins, ir, factory as f, helpers, types, visitor
from relationalai.early_access.metamodel.compiler import Pass, group_tasks
from relationalai.early_access.metamodel.util import OrderedSet, ordered_set
from relationalai.early_access.rel import metamodel_utils, dependency
from relationalai.early_access.metamodel.util import FrozenOrderedSet, filter_by_type

class Flatten(Pass):
    """
    Traverses the model's root to flatten it as much as possible. The result of this pass is
    a Logical root where all nested tasks that represent a rule in Rel are extraced to the
    top level.

    - nested logical with updates becomes a top-level logical (a rule)

    From:
        Logical
            Logical
                lookup1   <- scope is spread
                Logical
                    lookup2
                    derive foo
                Logical
                    lookup3
                    derive bar
    To:
        Logical
            Logical
                lookup1
                lookup2
                derive foo
            Logical
                lookup1
                lookup3
                derive bar

    - nested logical with aggregates becomes a top-level logical (a rule representing an aggregation)

    From:
        Logical
            Logical
                lookup1
                Logical
                    lookup2
                    aggregate1
                Logical
                    lookup3
                    aggregate2
                output
    To:
        Logical
            Logical
                lookup1
                lookup2
                aggregate1
                derive tmp1
            Logical
                lookup1
                lookup3
                aggregate2
                derive tmp2
            Logical
                lookup1
                lookup tmp1
                lookup tmp2
                output

    - a union becomes a top-level logical for each branch, writing into a temporary relation,
    and a lookup from that relation.

    From:
        Logical
            Logical
                Union
                    Logical
                        lookup1
                    Logical
                        lookup2
                output
    To:
        Logical
            Logical
                lookup1
                derive tmp1
            Logical
                lookup2
                derive tmp1
            Logical
                lookup tmp1
                output

    - a match becomes a top-level logical for each branch, each writing into its own temporary
    relation and a lookup from the last relation. The top-level logical for a branch derives
    into the temporary relation negating the previous branch:

    From:
        Logical
            Logical
                Match
                    Logical
                        lookup1
                    Logical
                        lookup2
                output
    To:
        Logical
            Logical
                lookup1
                derive tmp1
            Logical
                Union            <- tmp1() or (not temp1() and lookup2())
                    lookup tmp1
                    Logical
                        Not
                            lookup tmp1
                        lookup2
                        derive tmp2
            Logical
                lookup tmp2
                output
    """

    #--------------------------------------------------
    # Public API
    #--------------------------------------------------
    def rewrite(self, model: ir.Model, options:dict={}) -> ir.Model:
        # create the context
        ctx = Flatten.Context(model, options)

        # rewrite the root
        result = self.handle(model.root, ctx)

        # the new body contains the extracted top level logicals and maybe the rewritten root
        body = ctx.rewrite_ctx.top_level if result.replacement is None else ctx.rewrite_ctx.top_level + [result.replacement]

        # create the new model, updating relations and root
        return ir.Model(
            model.engines,
            OrderedSet.from_iterable(model.relations).update(ctx.rewrite_ctx.relations).frozen(),
            model.types,
            ir.Logical(model.root.engine, tuple(), tuple(body))
        )

    #--------------------------------------------------
    # IR handlers
    #--------------------------------------------------

    class Context():
        def __init__(self, model: ir.Model, options: dict):
            self.rewrite_ctx = helpers.RewriteContext()
            self.info = dependency.analyze(model.root)
            self.options = options
            # when we extract a nested logical (due to agg, match, union, etc), we also
            # extract together its dependencies, and we leave behind a different nested
            # logical as a reference, with a lookup to the extracted relation. This dict
            # keeps a map from the reference logical id to the set of tasks that where
            # extracted for it. This is used to optimize the resulting code, since we can
            # assume that the reference logical already includes those tasks.
            self.included: dict[int, OrderedSet[ir.Task]] = dict()

    @dataclass
    class HandleResult():
        replacement: Optional[ir.Task]
        updates: Optional[OrderedSet[ir.Update]] = None

    def handle(self, task: ir.Task, ctx: Context) -> Flatten.HandleResult:
        if isinstance(task, ir.Logical):
            return self.handle_logical(task, ctx)
        elif isinstance(task, ir.Union):
            return self.handle_union(task, ctx)
        elif isinstance(task, ir.Match):
            return self.handle_match(task, ctx)
        elif isinstance(task, ir.Require):
            return self.handle_require(task, ctx)
        else:
            return Flatten.HandleResult(task)

    def handle_logical(self, task: ir.Logical, ctx: Context):
        # keep track of what's the result of handling nested composites
        composites = group_tasks(task.body, {
            "composites": helpers.COMPOSITES
        })["composites"]
        all_composites_removed = len(composites) > 0

        # recursively handle children, collecting the replacements in the body
        body:OrderedSet[ir.Task] = ordered_set()
        for child in task.body:
            result = self.handle(child, ctx)
            if result.replacement is not None:
                ctx.info.replaced(child, result.replacement)
                extend_body(body, result.replacement)
                # nested composite was not completely removed
                if child in composites:
                    all_composites_removed = False

        # all children were extracted or all composites were removed without any effects
        # left and no outputs (so no way for outer dependencies), drop this logical
        if not body or (all_composites_removed and not any([isinstance(t, helpers.EFFECTS) for t in body]) and not ctx.info.task_outputs(task)):
            return Flatten.HandleResult(None)

        # now process the rewritten body
        groups = group_tasks(body.get_list(), {
            "outputs": ir.Output,
            "updates": ir.Update,
            "aggregates": ir.Aggregate,
            "ranks": ir.Rank,
        })

        # if there are outputs, adjust them (depending on the config for wide vs gnf)
        if groups["outputs"]:
            return self.adjust_outputs(task, body, groups, ctx)

        # if there are updates, extract as a new top level rule
        if groups["updates"]:
            # add dependencies for the logical, which is in the parent frame
            # if frame._parent_frame:
            #     body = frame._parent_frame.task_dependencies(task) | body
            body.prefix(ctx.info.task_dependencies(task))
            ctx.rewrite_ctx.top_level.append(ir.Logical(task.engine, task.hoisted, tuple(body), task.annotations))
            return Flatten.HandleResult(None)

        if groups["aggregates"]:
            if len(groups["aggregates"]) > 1:
                # stop rewritting as we don't know how to handle this yet
                return Flatten.HandleResult(task)

            # there must be only one
            agg = cast(ir.Aggregate, groups["aggregates"].some())

            # add agg dependencies to the body
            body.prefix(ctx.info.task_dependencies(agg))
            # body = frame.task_dependencies(agg) | body

            # extract a new logical for the aggregate, exposing aggregate group-by and results
            exposed_vars = OrderedSet.from_iterable(list(agg.group) + helpers.aggregate_outputs(agg))
            connection = metamodel_utils.extract(agg, body, exposed_vars.get_list(), ctx.rewrite_ctx)

            # return a reference to the connection relation
            reference = f.logical([f.lookup(connection, exposed_vars.get_list())], task.hoisted)
            return Flatten.HandleResult(reference)

        if groups["ranks"]:
            if len(groups["ranks"]) > 1:
                # stop rewritting as we don't know how to handle this yet
                return Flatten.HandleResult(task)

            # there must be only one
            rank = cast(ir.Rank, groups["ranks"].some())

            # add rank dependencies to the body
            # body = frame.task_dependencies(rank) | body
            body.prefix(ctx.info.task_dependencies(rank))

            # extract a new logical for the rank, exposing rank group-by and results
            exposed_vars = OrderedSet.from_iterable(list(rank.group) + list(rank.args) + [rank.result])
            connection = metamodel_utils.extract(rank, body, exposed_vars.get_list(), ctx.rewrite_ctx)

            # return a reference to the connection relation
            reference = f.logical([f.lookup(connection, exposed_vars.get_list())], task.hoisted)
            ctx.included[reference.id] = body
            return Flatten.HandleResult(reference)

        return Flatten.HandleResult(ir.Logical(task.engine, task.hoisted, tuple(body)))


    def handle_match(self, match: ir.Match, ctx: Context):
        # TODO: how to deal with malformed input like this?
        if not match.tasks:
            return Flatten.HandleResult(match)

        body = ctx.info.task_dependencies(match)
        outputs = ctx.info.task_outputs(match)
        exposed_vars = intersection(ctx.info.task_inputs(match), outputs)

        connection = None
        reference = None
        negation_len = len(outputs) if outputs else 0

        for branch in match.tasks:
            # process the branch
            result = self.handle(branch, ctx)
            assert(result.replacement)
            ctx.info.replaced(branch, result.replacement)

            branch_body: OrderedSet[ir.Task] = OrderedSet.from_iterable(body)
            extend_body(branch_body, result.replacement)
            if reference:
                branch_body.add(negate(reference, negation_len))
                branch_body = OrderedSet.from_iterable([f.union([f.logical(branch_body.get_list(), match.hoisted), reference], match.hoisted)])
            connection = metamodel_utils.extract(branch, branch_body, exposed_vars, ctx.rewrite_ctx, "match")
            reference = f.logical([f.lookup(connection, exposed_vars)], exposed_vars)

        return Flatten.HandleResult(reference)


    def handle_union(self, union: ir.Union, ctx: Context):
        # TODO: how to deal with malformed input like this?
        if not union.tasks:
            return Flatten.HandleResult(union)

        body = ctx.info.task_dependencies(union)
        outputs = ctx.info.task_outputs(union)
        exposed_vars = intersection(ctx.info.task_inputs(union), outputs)

        connection = None

        for branch in union.tasks:
            # process the branch
            result = self.handle(branch, ctx)
            if result.replacement:
                # the branch has some replacement, add it to the body to derive the union
                ctx.info.replaced(branch, result.replacement)
                branch_body: OrderedSet[ir.Task] = OrderedSet.from_iterable(body)
                extend_body(branch_body, result.replacement)
            else:
                # the branch was extracted as updates, generate a body that derives the
                # union when any of the updates is done
                # assert(result.updates)
                branch_body = ordered_set()
                if result.updates:
                    for update in result.updates:
                        branch_body.add(ir.Lookup(union.engine, update.relation, wildcards(update.relation)))

            if connection is None:
                # first branch, extract making a connection relation
                connection = metamodel_utils.extract(branch, branch_body, exposed_vars, ctx.rewrite_ctx, "union")
            else:
                # subsequent branch, extract reusing the connection relation
                # add derivation to the extracted body
                branch_body.add(f.derive(connection, exposed_vars))

                # extract the body
                ctx.rewrite_ctx.top_level.append(ir.Logical(union.engine, tuple(), tuple(branch_body)))

        # return a reference to the connection
        assert(connection)
        reference = f.logical([f.lookup(connection, exposed_vars)], exposed_vars)
        return Flatten.HandleResult(reference)

    def handle_require(self, req: ir.Require, ctx: Context):
        # only extract the domain if it is a somewhat complex Logical and there's more than
        # one check, otherwise insert it straight into all checks
        domain = req.domain
        if len(req.checks) > 1 and isinstance(domain, ir.Logical) and len(domain.body) > 1:
            body = OrderedSet.from_iterable(domain.body)
            vars = helpers.hoisted_vars(domain.hoisted)
            connection = metamodel_utils.extract(req, body, vars, ctx.rewrite_ctx, "domain")
            domain = f.logical([f.lookup(connection, vars)], vars)

        for check in req.checks:
            # only generate logic for checks that have errors
            if check.error:
                handled_check_result = self.handle(check.check, ctx)
                if handled_check_result.replacement:
                    body = ordered_set()
                    body.add(domain)
                    body.add(ir.Not(req.engine, handled_check_result.replacement))
                    if (isinstance(check.error, ir.Logical)):
                        body.update(check.error.body)
                    else:
                        # this is more general but may trip the current splinter
                        body.add(check.error)
                    ctx.rewrite_ctx.top_level.append(ir.Logical(req.engine, tuple(), tuple(body)))

        # currently we just drop the Require, but we should keep it here and link the
        # extracted logicals to it
        return Flatten.HandleResult(None)


    #--------------------------------------------------
    # GNF vs wide output support
    #--------------------------------------------------
    def adjust_outputs(self, task: ir.Logical, body: OrderedSet[ir.Task], groups: dict[str, OrderedSet[ir.Task]], ctx: Context):

        # for wide outputs, only adjust the output task to include the keys.
        if ctx.options.get("wide_outputs", False):
            for output in groups["outputs"]:
                assert(isinstance(output, ir.Output))
                if output.keys:
                    body.remove(output)
                    body.add(self.rewrite_wide_output(output))
            self.remove_subsumptions(body, ctx)
            return Flatten.HandleResult(ir.Logical(task.engine, task.hoisted, tuple(body), task.annotations))

        # for GNF outputs we need to generate a rule for each "column" in the output
        else:
            # first split outputs in potentially multiple outputs, one for each "column"
            for output in groups["outputs"]:
                assert(isinstance(output, ir.Output))
                if output.keys:
                    # we will replace the output bellow,
                    body.remove(output)

                    is_export = builtins.export_annotation in output.annotations
                    # generate an output for each "column"
                    for idx, alias in enumerate(output.aliases):
                        # output looks like def output(:cols, :col000, key0, key1, value):
                        aliases = []
                        if not is_export:
                            aliases.append(("cols", f.literal("cols", types.Symbol)))
                        aliases.append(("col", f.literal(f"col{idx:03}", types.Symbol)))
                        for key in output.keys:
                            aliases.append((f"key_{key.name}_{idx}", key))
                        aliases.append(alias)
                        body.add(ir.Output(
                            output.engine,
                            FrozenOrderedSet.from_iterable(aliases),
                            output.keys,
                            output.annotations
                        ))

            # analyse the resulting logical to be able to pull dependencies
            logical = ir.Logical(task.engine, task.hoisted, tuple(body), task.annotations)
            info = dependency.analyze(logical)

            # now extract a logical for each output, bringing together its dependencies
            for output in filter_by_type(body, ir.Output):
                deps = info.task_dependencies(output)
                # TODO: verify safety of doing this
                # self.remove_subsumptions(deps, ctx)
                deps.add(output)
                ctx.rewrite_ctx.top_level.append(ir.Logical(task.engine, tuple(), tuple(deps)))

            return Flatten.HandleResult(None)


    def remove_subsumptions(self, body:OrderedSet[ir.Task], ctx: Context):
        # remove from the body all the tasks that are subsumed by some other task in the set;
        # this can be done because some tasks are references to extracted nested logical that
        # contain filters they dependend on, so we don't need those filters here if the
        # reference is present.
        for logical in filter_by_type(body, ir.Logical):
            if logical.id in ctx.included:
                # if the logical id is included, it means it's a reference to an extracted
                # rule, so remove all other items in the body that are already included in
                # the body referenced by it
                for item in body:
                    if item in ctx.included[logical.id]:
                        body.remove(item)


    def rewrite_wide_output(self, output: ir.Output):
        assert(output.keys)

        # only prefix keys that are not already in the output
        prefix_keys = []
        for key in output.keys:
            if all([val is not key for _, val in output.aliases]):
                prefix_keys.append(key)

        aliases: OrderedSet[Tuple[str, ir.Value]] = ordered_set()
        # add the keys to the output
        for key in prefix_keys:
            aliases.add((key.name, key))

        # add the remaining args, unless it is already a key
        for name, val in output.aliases:
            if not isinstance(val, ir.Var) or val not in prefix_keys:
                aliases.add((name, val))

        # TODO - we are assuming that the Rel compiler will translate nullable lookups
        # properly, returning a `Missing` if necessary, like this:
        # (nested_192(_adult, _adult_name) or (not nested_192(_adult, _) and _adult_name = Missing)) and
        return ir.Output(
            output.engine,
            aliases.frozen(),
            output.keys,
            output.annotations
        )

        # TODO: in the rel compiler, see if we can do this outer join
        # 1. number of keys
        # 2. each relation
        # 3. each variable, starting with the keys
        # 4. tag output with @arrow

        # @arrow def output(_book, _book_title, _author_name):
        #   rel_primitive_outer_join(#1, book_title, author_name, _book, _book_title, _author_name)
        # def output(p, n, c):
        #     rel_primitive_outer_join(#1, name, coolness, p, n, c)

#--------------------------------------------------
# Helpers
#--------------------------------------------------

T = TypeVar('T')
def intersection(s1: Optional[OrderedSet[T]], s2: Optional[OrderedSet[T]]) -> list:
    """ Return the intersection of s1 | s2, accounting for when any is None. """
    if s1 and s2:
        return (s1 | s2).get_list()
    if s1:
        return s1.get_list()
    if s2:
        return s2.get_list()
    return []

def wildcards(relation: ir.Relation):
    return tuple([f.wild() for _ in relation.fields])

def extractable(t: ir.Task):
    """
    Whether this task is a Logical that will be extracted as a top level by this
    pass, because it has an aggregation, effects, match, union, etc.
    """
    extractable_types = (ir.Update, ir.Aggregate, ir.Match, ir.Union, ir.Rank)
    return isinstance(t, ir.Logical) and len(visitor.collect_by_type(extractable_types, t)) > 0

def extractables(composites: OrderedSet[ir.Task]):
    """ Filter the set of composites, keeping only the extractable ones. """
    return list(filter(extractable, composites))

def negate(reference: ir.Logical, values: int):
    """
    Return a negation of this reference, where the last `values` arguments are to
    be replaced by wildcards (i.e. len(reference.args) - values are keys so they need
    to be bound in the Not.)
    """
    lookup = cast(ir.Lookup, reference.body[0])
    args = []
    i = 0
    last = len(lookup.args) - values
    for arg in lookup.args:
        args.append(f.wild()) if i >= last else args.append(arg)
        i += 1

    return ir.Not(reference.engine, f.lookup(lookup.relation, args))

def merge_var_list(vars: list[ir.Var], hoisted: tuple[ir.VarOrDefault, ...]) -> list[ir.VarOrDefault]:
    """ Merge vars and hoisted, making sure that hoisted vars have precedence since they may have defaults. """
    r = []
    hoisted_vars = helpers.hoisted_vars(hoisted)
    for v in vars:
        if v not in hoisted_vars:
            r.append(v)
    r.extend(hoisted)
    return r

def extend_body(body: OrderedSet[ir.Task], extra: ir.Task):
    """ Add the extra task to the body, but if the extra is a simple logical, just
    inline its subtasks. """
    if isinstance(extra, ir.Logical):
        if extra.hoisted:
            # hoists, remove things that are already in the body to avoid duplicates
            logical_body = []
            for t in extra.body:
                if t not in body:
                    logical_body.append(t)
            if len(logical_body) == len(extra.body):
                # no duplicates
                body.add(extra)
            else:
                # some duplicate, remove them
                body.add(ir.Logical(
                    extra.engine,
                    extra.hoisted,
                    tuple(logical_body)
                ))
        else:
            # no hoists, just inline
            body.update(extra.body)
    else:
        body.add(extra)
